import { isBuiltin } from 'node:module';
import { IGNORE_DEFINITELY_TYPED, IGNORED_DEPENDENCIES } from './constants.js';
import { getPackageNameFromModuleSpecifier, isDefinitelyTyped, getDefinitelyTypedFor, getPackageFromDefinitelyTyped, } from './util/modules.js';
import { isAbsolute } from './util/path.js';
export class DependencyDeputy {
    isStrict;
    _manifests = new Map();
    referencedDependencies;
    peerDependencies;
    installedBinaries;
    constructor({ isStrict }) {
        this.isStrict = isStrict;
        this.referencedDependencies = new Map();
        this.peerDependencies = new Map();
        this.installedBinaries = new Map();
    }
    addWorkspace({ name, dir, manifestPath, manifest, ignoreDependencies, }) {
        const scripts = Object.values(manifest.scripts ?? {});
        const dependencies = Object.keys(manifest.dependencies ?? {});
        const peerDependencies = Object.keys(manifest.peerDependencies ?? {});
        const optionalDependencies = Object.keys(manifest.optionalDependencies ?? {});
        const devDependencies = Object.keys(manifest.devDependencies ?? {});
        const allDependencies = [...dependencies, ...devDependencies, ...peerDependencies, ...optionalDependencies];
        this._manifests.set(name, {
            workspaceDir: dir,
            manifestPath,
            ignoreDependencies,
            scripts,
            dependencies,
            peerDependencies,
            optionalDependencies,
            devDependencies,
            allDependencies,
        });
    }
    getWorkspaceManifest(workspaceName) {
        return this._manifests.get(workspaceName);
    }
    getProductionDependencies(workspaceName) {
        return this._manifests.get(workspaceName)?.dependencies ?? [];
    }
    getDevDependencies(workspaceName) {
        return this._manifests.get(workspaceName)?.devDependencies ?? [];
    }
    setInstalledBinaries(workspaceName, installedBinaries) {
        this.installedBinaries.set(workspaceName, installedBinaries);
    }
    getInstalledBinaries(workspaceName) {
        return this.installedBinaries.get(workspaceName);
    }
    addReferencedDependency(workspaceName, packageName) {
        if (!this.referencedDependencies.has(workspaceName)) {
            this.referencedDependencies.set(workspaceName, new Set());
        }
        this.referencedDependencies.get(workspaceName)?.add(packageName);
    }
    addPeerDependencies(workspaceName, peerDependencies) {
        this.peerDependencies.set(workspaceName, peerDependencies);
    }
    getPeerDependencies(workspaceName, dependency) {
        return Array.from(this.peerDependencies.get(workspaceName)?.get(dependency) ?? []);
    }
    maybeAddReferencedExternalDependency(workspace, moduleSpecifier) {
        if (this.isInternalDependency(moduleSpecifier))
            return [false];
        const packageName = getPackageNameFromModuleSpecifier(moduleSpecifier);
        const workspaceNames = this.isStrict ? [workspace.name] : [workspace.name, ...[...workspace.ancestors].reverse()];
        const closestWorkspaceName = workspaceNames.find(name => this.isInDependencies(name, packageName));
        if (this.getWorkspaceManifest(workspace.name)?.ignoreDependencies.includes(packageName))
            return [false];
        const typesPackageName = !isDefinitelyTyped(packageName) && getDefinitelyTypedFor(packageName);
        const closestWorkspaceNameForTypes = typesPackageName && workspaceNames.find(name => this.isInDependencies(name, typesPackageName));
        if (closestWorkspaceName || closestWorkspaceNameForTypes) {
            closestWorkspaceName && this.addReferencedDependency(closestWorkspaceName, packageName);
            closestWorkspaceNameForTypes && this.addReferencedDependency(closestWorkspaceNameForTypes, typesPackageName);
            return [true, packageName];
        }
        for (const name of workspaceNames) {
            const binaries = this.getInstalledBinaries(name);
            if (binaries?.has(moduleSpecifier)) {
                const dependencies = binaries.get(moduleSpecifier);
                if (dependencies?.size) {
                    dependencies.forEach(dependency => this.addReferencedDependency(name, dependency));
                    return [true, packageName];
                }
            }
        }
        return [false, moduleSpecifier];
    }
    isInternalDependency(moduleSpecifier) {
        return isAbsolute(moduleSpecifier) || moduleSpecifier.startsWith('.') || isBuiltin(moduleSpecifier);
    }
    isInDependencies(workspaceName, packageName) {
        const manifest = this._manifests.get(workspaceName);
        const dependencies = manifest ? (this.isStrict ? manifest.dependencies : manifest.allDependencies) : [];
        return dependencies.includes(packageName);
    }
    settleDependencyIssues() {
        const dependencyIssues = [];
        const devDependencyIssues = [];
        for (const [workspaceName, { manifestPath, ignoreDependencies }] of this._manifests.entries()) {
            const referencedDependencies = this.referencedDependencies.get(workspaceName);
            const isUnreferencedDependency = (dependency) => {
                if (referencedDependencies?.has(dependency))
                    return false;
                const [scope, typedDependency] = dependency.split('/');
                if (scope === '@types') {
                    const typedPackageName = getPackageFromDefinitelyTyped(typedDependency);
                    if (IGNORE_DEFINITELY_TYPED.includes(typedPackageName))
                        return false;
                    const peerDependencies = this.getPeerDependencies(workspaceName, typedPackageName);
                    if (peerDependencies.length) {
                        return !peerDependencies.find(peerDependency => !isUnreferencedDependency(peerDependency));
                    }
                    return !referencedDependencies?.has(typedPackageName);
                }
                if (!referencedDependencies?.has(dependency)) {
                    const peerDependencies = this.getPeerDependencies(workspaceName, dependency);
                    return !peerDependencies.find(peerDependency => !isUnreferencedDependency(peerDependency));
                }
                return false;
            };
            this.getProductionDependencies(workspaceName)
                .filter(symbol => !IGNORED_DEPENDENCIES.includes(symbol) && !ignoreDependencies.includes(symbol))
                .filter(isUnreferencedDependency)
                .forEach(symbol => dependencyIssues.push({ type: 'dependencies', filePath: manifestPath, symbol }));
            this.getDevDependencies(workspaceName)
                .filter(symbol => !IGNORED_DEPENDENCIES.includes(symbol) && !ignoreDependencies.includes(symbol))
                .filter(isUnreferencedDependency)
                .forEach(symbol => devDependencyIssues.push({ type: 'devDependencies', filePath: manifestPath, symbol }));
        }
        return { dependencyIssues, devDependencyIssues };
    }
}
