import path from 'node:path';
import { ProjectPrincipal } from './project-principal.js';
import { isAbsolute } from './util/path.js';
const mergePaths = (cwd, compilerOptions = {}, paths = {}) => {
    const mergedPaths = { ...compilerOptions.paths, ...paths };
    const baseUrl = compilerOptions.baseUrl ?? '.';
    compilerOptions.paths = Object.keys(mergedPaths).reduce((paths, key) => {
        paths[key] = mergedPaths[key].map(entry => (isAbsolute(entry) ? entry : path.join(cwd, baseUrl, entry)));
        return paths;
    }, {});
    return compilerOptions;
};
const hasDefaultBaseUrl = (compilerOptions) => !compilerOptions.baseUrl || /^\.\/?$/.test(compilerOptions.baseUrl);
export class PrincipalFactory {
    principals = new Set();
    getPrincipal({ cwd, compilerOptions, report, paths, compilers }) {
        compilerOptions = mergePaths(cwd, compilerOptions, paths);
        const principal = this.findReusablePrincipal(compilerOptions);
        if (principal) {
            this.linkPrincipal(principal, cwd, compilerOptions.paths);
            return principal.principal;
        }
        else {
            return this.addNewPrincipal({ cwd, compilerOptions, report, paths, compilers });
        }
    }
    findReusablePrincipal(compilerOptions) {
        if (!hasDefaultBaseUrl(compilerOptions))
            return;
        const workspacePaths = compilerOptions?.paths ? Object.keys(compilerOptions.paths) : [];
        const principal = Array.from(this.principals).find(principal => principal.isDefaultBaseUrl && workspacePaths.every(p => !principal.paths.has(p)));
        return principal;
    }
    linkPrincipal(principal, cwd, paths) {
        Object.keys(paths ?? {}).forEach(p => principal.paths.add(p));
        principal.principal.compilerOptions.paths = { ...principal.principal.compilerOptions.paths, ...paths };
        principal.cwds.add(cwd);
    }
    addNewPrincipal({ cwd, compilerOptions, report, compilers }) {
        const principal = new ProjectPrincipal({ cwd, compilerOptions, report, compilers });
        const paths = new Set(Object.keys(compilerOptions?.paths ?? {}));
        const isDefaultBaseUrl = hasDefaultBaseUrl(compilerOptions);
        compilerOptions.baseUrl = path.join(cwd, compilerOptions.baseUrl ?? '.');
        this.principals.add({ principal, cwds: new Set([cwd]), paths, isDefaultBaseUrl });
        return principal;
    }
    getPrincipals() {
        return Array.from(this.principals).map(p => p.principal);
    }
}
