import { isBuiltin } from 'node:module';
import ts from 'typescript';
import * as ast from './ast-helpers.js';
export const getImportsAndExports = (sourceFile, options) => {
    const internalImports = new Map();
    const externalImports = new Set();
    const unresolvedImports = new Set();
    const exports = new Map();
    const aliasedExports = {};
    const importedInternalSymbols = new Map();
    const addImport = ({ specifier, symbol, identifier = '__anonymous' }) => {
        if (isBuiltin(specifier))
            return;
        const module = sourceFile.resolvedModules?.get(specifier, undefined);
        if (module?.resolvedModule) {
            const filePath = module.resolvedModule.resolvedFileName;
            if (filePath) {
                if (module.resolvedModule.isExternalLibraryImport) {
                    if (module.resolvedModule.extension === '.d.ts') {
                        externalImports.add(specifier);
                    }
                    else {
                        externalImports.add(module.resolvedModule.packageId?.name ?? specifier);
                    }
                }
                else {
                    const isStar = identifier === '*';
                    const isReExported = Boolean(isStar && !symbol);
                    if (!internalImports.has(filePath)) {
                        internalImports.set(filePath, {
                            specifier,
                            isStar,
                            isReExported,
                            isReExportedBy: new Set(),
                            symbols: new Set(),
                        });
                    }
                    const internalImport = internalImports.get(filePath);
                    if (isReExported) {
                        internalImport.isReExported = isReExported;
                        internalImport.isReExportedBy.add(sourceFile.fileName);
                    }
                    if (isStar) {
                        internalImport.isStar = isStar;
                    }
                    if (!isStar) {
                        internalImport.symbols.add(identifier);
                    }
                    if (isStar && symbol) {
                        importedInternalSymbols.set(symbol, filePath);
                    }
                }
            }
        }
        else {
            unresolvedImports.add(specifier);
        }
    };
    const maybeAddNamespaceAccessAsImport = ({ namespace, member }) => {
        const symbol = sourceFile.locals?.get(namespace);
        if (symbol) {
            const importedSymbolFilePath = importedInternalSymbols.get(symbol);
            if (importedSymbolFilePath) {
                const internalImport = internalImports.get(importedSymbolFilePath);
                internalImport?.symbols.add(member);
            }
        }
    };
    const addExport = ({ node, identifier, type, pos, members }) => {
        if (options.skipExports)
            return;
        if (exports.has(identifier)) {
            const item = exports.get(identifier);
            exports.set(identifier, { ...item, node, type, pos, members });
        }
        else {
            exports.set(identifier, { node, type, pos, members });
        }
    };
    const addAliasedExport = (symbol, alias) => {
        aliasedExports[symbol] = aliasedExports[symbol] ?? [symbol];
        aliasedExports[symbol].push(alias);
    };
    const maybeAddAliasedExport = (node, alias) => {
        if (node && ts.isIdentifier(node) && sourceFile.symbol?.exports?.has(node.getText())) {
            addAliasedExport(node.getText(), alias);
        }
    };
    const visit = (node) => {
        if (ts.isImportDeclaration(node) && ts.isStringLiteralLike(node.moduleSpecifier)) {
            const specifier = node.moduleSpecifier.text;
            if (!node.importClause) {
                addImport({ specifier });
            }
            else {
                if (node.importClause.isTypeOnly && options.skipTypeOnly)
                    return;
                if (ast.isDefaultImport(node)) {
                    addImport({ specifier, identifier: 'default' });
                }
                if (node.importClause?.namedBindings) {
                    if (ts.isNamespaceImport(node.importClause.namedBindings)) {
                        const symbol = node.importClause.namedBindings.symbol;
                        addImport({ symbol, specifier, identifier: '*' });
                    }
                    if (ts.isNamedImports(node.importClause.namedBindings)) {
                        node.importClause.namedBindings.elements.forEach(element => {
                            const identifier = (element.propertyName ?? element.name).getText();
                            addImport({ symbol: element.symbol, specifier, identifier });
                        });
                    }
                }
            }
        }
        if (ast.isImportCall(node)) {
            if (node.arguments[0] && ts.isStringLiteralLike(node.arguments[0])) {
                const specifier = node.arguments[0].text;
                let _node = node.parent;
                while (_node) {
                    if (ts.isExpressionStatement(_node)) {
                        addImport({ specifier });
                        break;
                    }
                    if (_node.parent && ts.isCallExpression(_node.parent)) {
                        addImport({ specifier, identifier: 'default' });
                        break;
                    }
                    if (ast.isAccessExpression(_node)) {
                        const identifier = ast.getAccessExpressionName(_node);
                        const isPromiseLike = identifier === 'then';
                        const symbol = isPromiseLike ? 'default' : identifier;
                        addImport({ identifier: symbol, specifier });
                        break;
                    }
                    if (ast.isVariableDeclarationList(_node)) {
                        const variableDeclarations = ast.findDescendants(_node, _node => ts.isVariableDeclaration(_node));
                        variableDeclarations.forEach(variableDeclaration => {
                            if (ts.isIdentifier(variableDeclaration.name)) {
                                addImport({ identifier: 'default', specifier });
                            }
                            else {
                                const binds = ast.findDescendants(variableDeclaration, _node => ts.isBindingElement(_node));
                                binds.forEach(element => {
                                    const symbol = element.propertyName?.getText() || element.name.getText();
                                    addImport({ identifier: symbol, specifier });
                                });
                            }
                        });
                        break;
                    }
                    _node = _node.parent;
                }
            }
        }
        if (ts.isImportEqualsDeclaration(node) &&
            ts.isExternalModuleReference(node.moduleReference) &&
            ts.isStringLiteralLike(node.moduleReference.expression)) {
            const specifier = node.moduleReference.expression.text;
            addImport({ specifier, identifier: 'default' });
        }
        if (ast.isRequireCall(node)) {
            if (ts.isStringLiteralLike(node.arguments[0])) {
                const specifier = node.arguments[0].text;
                if (specifier) {
                    const propertyAccessExpression = ast.findAncestor(node, _node => {
                        if (ts.isExpressionStatement(_node) || ts.isCallExpression(_node))
                            return 'STOP';
                        return ts.isPropertyAccessExpression(_node);
                    });
                    if (propertyAccessExpression) {
                        const identifier = String(propertyAccessExpression.name.escapedText);
                        addImport({ identifier, specifier });
                    }
                    else {
                        const variableDeclaration = node.parent;
                        if (ts.isVariableDeclaration(variableDeclaration) &&
                            ts.isVariableDeclarationList(variableDeclaration.parent)) {
                            if (ts.isIdentifier(variableDeclaration.name)) {
                                addImport({ identifier: 'default', specifier });
                            }
                            else {
                                const binds = ast.findDescendants(variableDeclaration, _node => ts.isBindingElement(_node));
                                if (binds.length > 0) {
                                    binds.forEach(element => {
                                        const identifier = (element.propertyName ?? element.name).getText();
                                        addImport({ identifier, specifier });
                                    });
                                }
                                else {
                                    addImport({ identifier: 'default', specifier });
                                }
                            }
                        }
                        else {
                            addImport({ identifier: 'default', specifier });
                        }
                    }
                }
            }
        }
        if (ast.isRequireResolveCall(node)) {
            if (node.arguments[0] && ts.isStringLiteralLike(node.arguments[0])) {
                const specifier = node.arguments[0].text;
                if (specifier) {
                    addImport({ specifier });
                }
            }
        }
        const modifierKinds = node.modifiers?.map(modifier => modifier.kind) ?? [];
        if (ts.isExportAssignment(node)) {
            addExport({ node, identifier: 'default', type: 'unknown', pos: node.expression.getStart() });
            maybeAddAliasedExport(node.expression, 'default');
        }
        if (modifierKinds.includes(ts.SyntaxKind.ExportKeyword)) {
            if (ts.isVariableStatement(node)) {
                node.declarationList.declarations.forEach(declaration => {
                    if (ts.isObjectBindingPattern(declaration.name)) {
                        declaration.name.elements.forEach(element => {
                            if (ts.isIdentifier(element.name)) {
                                addExport({
                                    node: element,
                                    identifier: element.name.escapedText.toString(),
                                    type: 'unknown',
                                    pos: element.name.getStart(),
                                });
                            }
                        });
                    }
                    else if (ts.isArrayBindingPattern(declaration.name)) {
                        declaration.name.elements.forEach(element => {
                            if (ts.isBindingElement(element)) {
                                addExport({
                                    node: element,
                                    identifier: element.getText(),
                                    type: 'unknown',
                                    pos: element.getStart(),
                                });
                            }
                        });
                    }
                    else {
                        const identifier = declaration.name.getText();
                        addExport({ node: declaration, identifier, type: 'unknown', pos: declaration.name.getStart() });
                        maybeAddAliasedExport(declaration.initializer, identifier);
                    }
                });
            }
            if (ts.isFunctionDeclaration(node) && node.name) {
                const identifier = modifierKinds.includes(ts.SyntaxKind.DefaultKeyword) ? 'default' : node.name.getText();
                const pos = (node.name ?? node.body ?? node).getStart();
                addExport({ node, identifier, pos, type: 'function' });
            }
            if (ts.isClassDeclaration(node) && node.name) {
                const identifier = modifierKinds.includes(ts.SyntaxKind.DefaultKeyword) ? 'default' : node.name.getText();
                const pos = (node.name ?? node).getStart();
                const members = node.members
                    .filter((member) => (ts.isPropertyDeclaration(member) || ts.isMethodDeclaration(member)) && !ast.isPrivateMember(member))
                    .map(n => ({ node: n, identifier: n.name.getText(), pos: n.name.getStart(), type: 'member' }));
                addExport({ node, identifier, type: 'class', pos, members });
            }
            if (ts.isTypeAliasDeclaration(node)) {
                addExport({ node, identifier: node.name.getText(), type: 'type', pos: node.name.getStart() });
            }
            if (ts.isInterfaceDeclaration(node)) {
                addExport({ node, identifier: node.name.getText(), type: 'interface', pos: node.name.getStart() });
            }
            if (ts.isEnumDeclaration(node)) {
                const identifier = modifierKinds.includes(ts.SyntaxKind.DefaultKeyword) ? 'default' : node.name.getText();
                const pos = node.name.getStart();
                const members = node.members.map(n => ({
                    node: n,
                    identifier: ast.stripQuotes(n.name.getText()),
                    pos: n.name.getStart(),
                    type: 'member',
                }));
                addExport({ node, identifier, type: 'enum', pos, members });
            }
        }
        if (ts.isExportDeclaration(node)) {
            if (node.moduleSpecifier && ts.isStringLiteralLike(node.moduleSpecifier)) {
                if (!node.exportClause) {
                    addImport({ identifier: '*', specifier: node.moduleSpecifier.text });
                }
                else if (node.exportClause.kind === ts.SyntaxKind.NamespaceExport) {
                    addImport({ identifier: '*', specifier: node.moduleSpecifier.text });
                }
                else {
                    const specifier = node.moduleSpecifier;
                    node.exportClause.elements.forEach(element => {
                        const identifier = (element.propertyName ?? element.name).getText();
                        addImport({ identifier, specifier: specifier.text });
                    });
                }
            }
            else if (node.exportClause && ts.isNamedExports(node.exportClause)) {
                const type = node.isTypeOnly ? 'type' : 'unknown';
                node.exportClause.elements.forEach(element => {
                    addExport({ node: element, identifier: element.name.getText(), type, pos: element.name.pos });
                });
            }
        }
        if (ast.isModuleExportsAccessExpression(node)) {
            const parent = node.parent;
            if (ts.isPropertyAccessExpression(parent)) {
                const identifier = parent.name.getText();
                const pos = parent.name.getStart();
                addExport({ node, identifier, type: 'unknown', pos });
            }
            else if (ts.isElementAccessExpression(parent)) {
                const identifier = ast.stripQuotes(parent.argumentExpression.getText());
                const pos = parent.argumentExpression.getStart();
                addExport({ node, identifier, type: 'unknown', pos });
            }
            else if (ts.isBinaryExpression(parent)) {
                const expr = parent.right;
                if (ts.isObjectLiteralExpression(expr) && expr.properties.every(ts.isShorthandPropertyAssignment)) {
                    expr.properties.forEach(node => {
                        addExport({ node, identifier: node.getText(), type: 'unknown', pos: node.pos });
                    });
                }
                else {
                    addExport({ node, identifier: 'default', type: 'unknown', pos: node.getStart() });
                }
            }
        }
        if ('jsDoc' in node) {
            const type = ts.getJSDocType(node);
            if (type && ast.isValidImportTypeNode(type)) {
                addImport({ specifier: type.argument.literal.text });
            }
        }
        if (ast.isAccessExpression(node)) {
            maybeAddNamespaceAccessAsImport({
                namespace: node.expression.getText(),
                member: ast.getAccessExpressionName(node),
            });
        }
        ts.forEachChild(node, visit);
    };
    visit(sourceFile);
    const duplicateExports = Object.values(aliasedExports);
    return { internalImports, externalImports, unresolvedImports, exports, duplicateExports };
};
